from . import MoGit
from .repo import Config
import click
import os


@click.group()
def cli():
    pass


@cli.command(help="初始化仓库")
def init():
    if os.path.exists(".mogit"):
        print("已初始化")
        return
    MoGit.init()
    print("初始化完毕 ")


@cli.command(help="加入暂存区")
@click.argument("path")
def add(path):
    mogit = MoGit()
    if os.path.exists(path):
        mogit.add(path)
    else:
        print("文件或目录不存在")


@cli.command(help="创建提交")
@click.argument("msg")
def commit(msg):
    mogit = MoGit()
    mogit.commit(msg)


@cli.command(help="检出提交")
@click.argument("idx")
def checkout(idx):
    mogit = MoGit()
    mogit.checkout(idx)


@cli.command(help="创建提交")
@click.argument("tag")
@click.option("-h", "--hash", default=False)
def tag(tag, hash):
    mogit = MoGit()
    if hash:
        mogit.tag(tag, hash)
    else:
        mogit.tag(tag, mogit.config.get("LastCommit"))


@cli.command(help="列出所有提交")
def log():
    mogit = MoGit()
    mogit.log()


@cli.command(help="分支操作")
@click.argument("opcode")
@click.option("-n", "--new")
def branch(opcode, new):
    mogit = MoGit()
    if new:
        mogit.create_branch(opcode)
    else:
        mogit.switch_branch(opcode)


@cli.command(help="设置")
@click.argument("key")
@click.argument("val")
def set(key, val):
    config = Config()
    config.set(key, val)


if __name__ == "__main__":
    cli()



from hashlib import sha1
import pickle
import os


class MoGitObj:
    def __init__(self, type, data):
        self.type = type
        self.data = data
        self.hash = ""
        self.calc_hash()

    def calc_hash(self):
        data = {"type": self.type, "data": self.data}
        byte = pickle.dumps(data)
        hasher = sha1(byte)
        self.hash = hasher.hexdigest()

    def save(self, storage_path):
        os.makedirs(
            os.path.join(storage_path, self.hash[:2], self.hash[2:4]), exist_ok=True
        )
        with open(
            os.path.join(storage_path, self.hash[:2], self.hash[2:4], self.hash), "wb"
        ) as f:
            pickle.dump(self, f)

    @staticmethod
    def load(storage_path, hash):
        path = os.path.join(storage_path, hash[:2], hash[2:4], hash)
        if os.path.exists(path):
            with open(path, "rb") as f:
                return pickle.load(f)

        return None


class Blob(MoGitObj):
    def __init__(self, content):
        super().__init__("blob", content)


class Commit(MoGitObj):
    def __init__(self, commit_data):
        super().__init__("commit", commit_data)

    def get_commit(self):
        return self.data


class MoGitStorage:
    def __init__(self, storage_path):
        self.storage_path = storage_path

    def add_object(self, obj):
        obj.save(self.storage_path)

    def get_object(self, hash):
        return MoGitObj.load(self.storage_path, hash)



import os
import json
from .objs import Blob
from pathlib import Path


class StagingArea:
    def __init__(self):
        self.path = os.path.join(".mogit", "staging.json")

    @staticmethod
    def init():
        path = os.path.join(".mogit", "staging.json")

        if not os.path.exists(path):
            with open(path, "w") as f:
                json.dump([], f, indent=4)

    def add(self, filename, objs):

        with open(filename, "rb") as f:
            data = {"name": filename, "content": f.read()}
            blob = Blob(data)
        objs.add_object(blob)
        datas = json.load(open(self.path))
        datas.append({"name": filename, "hash": blob.hash})
        self.write(datas)

    def write(self, data):
        with open(self.path, "w") as f:
            json.dump(data, f, indent=4)

    def clean(self):
        self.write([])

    def get_all(self):
        return json.load(open(self.path))


class RepoData:
    def __init__(self):
        self.path = os.path.join(".mogit", "data.json")

    @staticmethod
    def init():
        path = os.path.join(".mogit", "data.json")
        with open(path, "w") as f:
            json.dump({"Branch": "main"}, f, indent=4)

    def get(self, name):
        with open(self.path) as f:
            data = json.load(f)
            return data.get(name, None)

    def set(self, key, val):
        with open(self.path) as f:
            data = json.load(f)
            data[key] = val

        with open(self.path, "w") as f:
            json.dump(data, f, indent=4)


class CommitLog(StagingArea):
    def __init__(self, config):
        self.config = config

    def get_path(self, branch=None):
        if branch == None:
            branch = self.config.get("Branch")
        path = os.path.join(".mogit", "commits", f"{branch}.json")
        if not os.path.exists(path):
            os.makedirs(os.path.dirname(path), exist_ok=True)
            with open(path, "w") as f:
                json.dump([], f, indent=4)
        return path

    def add(self, hashs):
        datas = json.load(open(self.get_path()))
        datas.append(hashs)
        self.write(datas)

    def write(self, data):
        with open(self.get_path(), "w") as f:
            json.dump(data, f, indent=4)

    def get_all(self):
        return json.load(open(self.get_path()))

    def get(self, n):
        return json.load(open(self.get_path()))[n]


class Tags(RepoData):
    def __init__(self):
        self.path = os.path.join(".mogit", "tags.json")

    @staticmethod
    def init():
        path = os.path.join(".mogit", "tags.json")
        with open(path, "w") as f:
            json.dump({}, f, indent=4)


class Config(RepoData):
    def __init__(self):
        self.path = os.path.join(Path.home(), ".mogitconfig")

        if not os.path.exists(self.path):
            with open(self.path, "w") as f:
                json.dump({}, f, indent=4)

    def get(self, name):
        with open(self.path) as f:
            data = json.load(f)
            return data.get(name, None)

    def set(self, key, val):
        with open(self.path) as f:
            data = json.load(f)
            data[key] = val

        with open(self.path, "w") as f:
            json.dump(data, f, indent=4, ensure_ascii=False)



from setuptools import setup, find_packages

setup(
    name="mogit",  # 包的名称
    version="0.0.3",  # 版本号
    packages=find_packages(exclude=["tests", "tests.*"]),  # 包含的包
    author="MoYan",  # 作者
    author_email="moyan@moyanjdc.top",  # 作者邮箱
    description="一个轻量级的分布式版本控制系统",  # 包的简要描述
    long_description=open("Readme.md").read(),  # 包的详细描述
    long_description_content_type="text/markdown",  # 描述的内容类型
    classifiers=[  # 包的分类信息
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Environment :: Console",
        "Programming Language :: Python :: 3.10",
    ],
    install_requires=open("requirements.txt").read().split("\n"),
    entry_points={
        "console_scripts": ["mogit=mogit.cli:cli"],
    },
)



import os
from hashlib import sha1
import time

from .objs import MoGitStorage, Commit
from .repo import StagingArea, RepoData, CommitLog, Tags, Config
from . import ignorefile


class MoGit:
    def __init__(self):
        self.base_dir = ".mogit"
        self.objs = MoGitStorage(os.path.join(self.base_dir, "objs"))  # 源文件
        self.staging_area = StagingArea()
        self.config = RepoData()
        self.app_config = Config()
        self.commits = CommitLog(self.config)
        self.tags = Tags()

        if not os.path.exists(self.base_dir):
            print("MoGit未初始化")
            exit()

        if os.path.exists(".mogitignore"):
            with open(".mogitignore") as f:
                self.matchs = ignorefile.parse(f, ".")
        else:
            self.matchs = self.fake_match

    @staticmethod
    def init():
        """
        初始化仓库
        """
        base_dir = ".mogit"
        objs = os.path.join(base_dir, "objs")  # 源文件

        os.makedirs(objs, exist_ok=True)
        RepoData.init()
        StagingArea.init()
        CommitLog.init()
        Tags.init()

    def get_hash(self, hashs):
        """
        获取tag的hash
        """
        res = self.tags.get(hashs)
        if res is None:
            return hashs
        else:
            return res

    def hash_file(self, file_path):
        """
        计算文件hash
        """
        hasher = sha1()
        with open(file_path, "rb") as f:
            buf = f.read()
            hasher.update(buf)
        return hasher.hexdigest()

    def fake_match(self, path):
        return False

    def check_path(self, path):
        """
        检查
        """
        if not self.matchs(path) and ".mogit" + os.sep not in path:
            return True

        return False

    def add(self, file_path):
        """
        添加文件至暂存区
        """
        if os.path.isfile(file_path):
            self.staging_area.add(file_path, self.objs)

        elif os.path.isdir(file_path):
            for path, dirs, files in os.walk(file_path):
                for file in files:
                    file_path = os.path.join(path, file)

                    if self.check_path(file_path):
                        self.staging_area.add(file_path, self.objs)

    def commit(self, message):
        """
        提交文件
        """
        author = self.app_config.get("author")
        mail = self.app_config.get("mail")
        if author is None:
            print('WARNING: 未定义作者，可使用 "mogit set author <author>" 来设定')
        if author is None:
            print('WARNING: 未定义作者，可使用 "mogit set mail <mail>" 来设定')

        commit_data = {
            "timestamp": time.time(),
            "msg": message,
            "files": self.staging_area.get_all(),
            "author": author,
            "mail": mail,
        }

        self.staging_area.clean()

        comm = Commit(commit_data)
        self.objs.add_object(comm)
        self.commits.add(comm.hash)

        self.config.set("LastCommit", comm.hash)

    def checkout(self, hashs):
        """
        检出提交
        """
        obj = self.objs.get_object(self.get_hash(hashs))
        if obj is None:
            print("没有这个提交")
            exit()
        for path, dirs, files in os.walk("."):
            for file in files:
                file_path = os.path.join(path, file)

                if self.check_path(file_path):
                    os.remove(file_path)

        for i in obj.data["files"]:
            file = self.objs.get_object(i["hash"])
            with open(i["name"], "wb") as f:
                f.write(file.data["content"])
        self.config.set("LastCommit", hashs)

    def tag(self, tag, hashs):
        """
        提交文件
        """
        self.tags.set(tag, hashs)

    def log(self):
        """
        获取所有提交
        """
        for commit_hash in self.commits.get_all():
            print(f"Commit: {commit_hash}")

    def create_branch(self, name):
        all_commit = self.commits.get_all()
        self.config.set("Branch", name)
        self.commits.write(all_commit)

    def switch_branch(self, name):
        # 切换分支
        path = os.path.join(".mogit", "commits", f"{name}.json")

        if os.path.exists(path):
            self.config.set("Branch", name)
        else:
            print("分支不存在")
        last_commit = self.commits.get(-1)
        self.config.set("LastCommit", last_commit)
        if last_commit:
            self.checkout(last_commit)


if __name__ == "__main__":

    # 使用示例
    MoGit.init()
    simple_git = MoGit()
    simple_git.create_branch("")
    for i in range(4):
        simple_git.add(".")
        simple_git.commit("Initial commit")

    simple_git.tag("a", "93148e363210a8c3f8bdfac9c544d87c8cf8f6be")



from flask import Flask, request
from . import MoGit
import os

app = Flask(__file__)


@app.route("/clone/<repo>")
def repo_metadata(repo):
    base_dir = app.config.get("repo_base_dir")
    repo_dir = os.path.join(base_dir, repo)

    if not os.path.exists(repo_dir):
        return {"msg": "仓库不存在"}, 404

    old_dir = os.getcwd()
    os.chdir(repo_dir)
    mogit = MoGit()
    last_commit = mogit.config.get("LastCommit")
    commits = mogit.commits.get_all()
    branchs = []
    for f in os.listdir(".mogit/commits"):
        name, ext = os.path.splitext(f)
        branchs.append(name)

    data = {
        "last": last_commit,
        "commits": commits,
        "branchs": branchs,
        "branch": mogit.config.get("Branch"),
    }
    os.chdir(old_dir)
    return data


@app.route("/clone/<repo>/list")
def clone_list(repo):
    base_dir = app.config.get("repo_base_dir")
    repo_dir = os.path.join(base_dir, repo)
    commit = request.args.get("commit", "l")
    branch = request.args.get("branch", "d")
    if not os.path.exists(repo_dir):
        return {"msg": "仓库不存在"}, 404

    old_dir = os.getcwd()
    os.chdir(repo_dir)
    mogit = MoGit()

    return "尚未实现"


if __name__ == "__main__":
    app.config["repo_base_dir"] = "."
    app.run()



"""A spec-compliant `.gitignore` parser for Python."""

import collections
import os
import re


DEFAULT_IGNORE_NAMES = [".mogitignore"]


def parse(path, base_path=None):
    """Parses single `.gitignore` file.

    Args:
        path (str): Path to `.gitignore` file.
        base_path (str): Base path for applying ignore rules.

    Returns:
        Callable[[str], bool]: Callable which returns `True` if specified path is ignored.
            You can also pass `is_dir: bool` optional parameter if you know whether the specified path is a directory.
    """

    if base_path is None:
        base_path = os.path.dirname(path) or os.path.dirname(os.path.abspath(path))

    rules = []
    # with open(path) as ignore_file:
    for line in path:
        line = line.rstrip("\r\n")
        rule = _rule_from_pattern(line)
        if rule:
            rules.append(rule)

    return _IgnoreRules(rules, base_path).match


def ignore(ignore_names=DEFAULT_IGNORE_NAMES):
    """Returns `shutil.copytree()`-compatible ignore function for skipping ignored files.

    It will check if file is ignored by any `.gitignore` in the directory tree.

    Args:
        ignore_names (list[str], optional): List of names of ignore files.

    Returns:
        Callable[[str, list[str]], list[str]]: Callable compatible with `shutil.copytree()`.
    """

    matches = Cache(ignore_names=ignore_names)
    return lambda root, names: {
        name for name in names if matches(os.path.join(root, name))
    }


def ignored(path, is_dir=None, ignore_names=DEFAULT_IGNORE_NAMES):
    """Checks if file is ignored by any `.gitignore` in the directory tree.

    Args:
        path (str): Path to check against ignore rules.
        is_dir (bool, optional): Set if you know whether the specified path is a directory.
        ignore_names (list[str], optional): List of names of ignore files.

    Returns:
        bool: `True` if the path is ignored.
    """

    return Cache(ignore_names=ignore_names)(path, is_dir=is_dir)


class Cache:
    """Caches information about different `.gitignore` files in the directory tree.

    Allows to reduce number of queries to filesystem to mininum.
    """

    def __init__(self, ignore_names=DEFAULT_IGNORE_NAMES):
        """Constructs `Cache` objects.

        Args:
            ignore_names (list[str], optional): List of names of ignore files.
        """

        self.__ignore_names = ignore_names
        self.__gitignores = {}

    def __call__(self, path, is_dir=None):
        """Checks whether the specified path is ignored.

        Args:
            path (str): Path to check against ignore rules.
            is_dir (bool, optional): Set if you know whether the specified path is a directory.
        """

        path = _Path(path)
        add_to_children = {}
        plain_paths = []
        for parent in path.parents():
            if parent.parts in self.__gitignores:
                break

            ignore_paths = []
            for ignore_name in self.__ignore_names:
                ignore_path = parent.join(ignore_name)
                if ignore_path.isfile():
                    ignore_paths.append(str(ignore_path))

            if ignore_paths:
                matches = [
                    parse(ignore_path, base_path=parent) for ignore_path in ignore_paths
                ]
                add_to_children[parent] = (matches, plain_paths)
                plain_paths = []

            else:
                plain_paths.append(parent)

        else:
            parent = _Path(tuple())  # Null path.
            self.__gitignores[parent.parts] = []

        for plain_path in plain_paths:
            # assert plain_path.parts not in self.__gitignores
            self.__gitignores[plain_path.parts] = self.__gitignores[parent.parts]

        for parent, (_, parent_plain_paths) in reversed(list(add_to_children.items())):
            # assert parent.parts not in self.__gitignores
            self.__gitignores[parent.parts] = self.__gitignores[
                parent.parts[:-1]
            ].copy()
            for parent_to_add, (gitignores_to_add, _) in reversed(
                list(add_to_children.items())
            ):
                self.__gitignores[parent.parts].extend(gitignores_to_add)
                if parent_to_add == parent:
                    break

            self.__gitignores[parent.parts].reverse()

            for plain_path in parent_plain_paths:
                # assert plain_path.parts not in self.__gitignores
                self.__gitignores[plain_path.parts] = self.__gitignores[parent.parts]

        # This parent comes either from first or second loop.
        return any((m(path, is_dir=is_dir) for m in self.__gitignores[parent.parts]))


class _Path:
    def __init__(self, path):
        if isinstance(path, str):
            abs_path = os.path.abspath(path)
            self.__parts = tuple(_path_split(abs_path))
            self.__joined = abs_path
            self.__is_dir = None

        else:
            self.__parts = path
            self.__joined = None
            self.__is_dir = None

    @property
    def parts(self):
        return self.__parts

    def join(self, name):
        return _Path(self.__parts + (name,))

    def relpath(self, base_path):
        if self.__parts[: len(base_path.__parts)] == base_path.__parts:
            return "/".join(self.__parts[len(base_path.__parts) :])

        else:
            return None

    def parents(self):
        for i in range(len(self.__parts) - 1, 0, -1):
            yield _Path(self.__parts[:i])

    def isfile(self):
        return os.path.isfile(str(self))

    def isdir(self):
        if self.__is_dir is not None:
            return self.__is_dir
        self.__is_dir = os.path.isdir(str(self))
        return self.__is_dir

    def __str__(self):
        if self.__joined is None:
            self.__joined = (
                os.sep.join(self.__parts) if self.__parts != ("",) else os.sep
            )
        return self.__joined


def _rule_from_pattern(pattern):
    # Takes a `.gitignore` match pattern, such as "*.py[cod]" or "**/*.bak",
    # and returns an `_IgnoreRule` suitable for matching against files and
    # directories. Patterns which do not match files, such as comments
    # and blank lines, will return `None`.

    # Store the exact pattern for our repr and string functions
    orig_pattern = pattern

    # Early returns follow
    # Discard comments and separators
    if not pattern.lstrip() or pattern.lstrip().startswith("#"):
        return

    # Discard anything with more than two consecutive asterisks
    if "***" in pattern:
        return

    # Strip leading bang before examining double asterisks
    if pattern.startswith("!"):
        negation = True
        pattern = pattern[1:]
    else:
        negation = False

    # Discard anything with invalid double-asterisks -- they can appear
    # at the start or the end, or be surrounded by slashes
    for m in re.finditer("\\*\\*", pattern):
        start_index = m.start()
        if (
            start_index != 0
            and start_index != len(pattern) - 2
            and (pattern[start_index - 1] != "/" or pattern[start_index + 2] != "/")
        ):
            return

    # Special-casing '/', which doesn't match any files or directories
    if pattern.rstrip() == "/":
        return

    directory_only = pattern.endswith("/")

    # A slash is a sign that we're tied to the `base_path` of our rule
    # set.
    anchored = "/" in pattern[:-1]

    if pattern.startswith("/"):
        pattern = pattern[1:]
    if pattern.startswith("**"):
        pattern = pattern[2:]
        anchored = False
    if pattern.startswith("/"):
        pattern = pattern[1:]
    if pattern.endswith("/"):
        pattern = pattern[:-1]

    # patterns with leading hashes are escaped with a backslash in front, unescape it
    if pattern.startswith("\\#"):
        pattern = pattern[1:]

    # trailing spaces are ignored unless they are escaped with a backslash
    i = len(pattern) - 1
    striptrailingspaces = True
    while i > 1 and pattern[i] == " ":
        if pattern[i - 1] == "\\":
            pattern = pattern[: i - 1] + pattern[i:]
            i -= 1
            striptrailingspaces = False
        else:
            if striptrailingspaces:
                pattern = pattern[:i]
        i -= 1

    regexp = _fnmatch_pathname_to_regexp(pattern, anchored, directory_only)
    return _IgnoreRule(regexp, negation, directory_only)


class _IgnoreRules:
    def __init__(self, rules, base_path):
        self.__rules = rules
        self.__can_return_immediately = not any((r.negation for r in rules))
        self.__base_path = _Path(base_path) if isinstance(base_path, str) else base_path

    def match(self, path, is_dir=None):
        if isinstance(path, str):
            path = _Path(path)

        rel_path = path.relpath(self.__base_path)

        if rel_path is not None:
            if is_dir is None:
                is_dir = path.isdir()  # TODO Pass callable here.

            if self.__can_return_immediately:
                return any((r.match(rel_path, is_dir) for r in self.__rules))

            else:
                matched = False
                for rule in self.__rules:
                    if rule.match(rel_path, is_dir):
                        matched = not rule.negation

                else:
                    return matched

        else:
            return False


class _IgnoreRule:
    def __init__(self, regexp, negation, directory_only):
        self.__regexp = re.compile(regexp)
        self.__negation = negation
        self.__directory_only = directory_only
        self.__match = self.__regexp.match

    @property
    def regexp(self):
        return self.__regexp

    @property
    def negation(self):
        return self.__negation

    def match(self, rel_path, is_dir):
        m = self.__match(rel_path)

        # If we need a directory, check there is something after slash and if there is not, target must be a directory.
        # If there is something after slash then it's a directory irrelevant to type of target.
        # `self.directory_only` implies we have group number 1.
        # N.B. Question mark inside a group without a name can shift indices. :(
        return m and (not self.__directory_only or m.group(1) is not None or is_dir)


if os.altsep is not None:
    _all_seps_expr = f"[{re.escape(os.sep)}{re.escape(os.altsep)}]"
    _path_split = lambda path: re.split(_all_seps_expr, path)

else:
    _path_split = lambda path: path.split(os.sep)


def _fnmatch_pathname_to_regexp(pattern, anchored, directory_only):
    # Implements `fnmatch` style-behavior, as though with `FNM_PATHNAME` flagged;
    # the path separator will not match shell-style `*` and `.` wildcards.

    # Frustratingly, python's fnmatch doesn't provide the FNM_PATHNAME
    # option that `.gitignore`'s behavior depends on.

    if not pattern:
        if directory_only:
            return "[^/]+(/.+)?$"  # Empty name means no path fragment.

        else:
            return ".*"

    i, n = 0, len(pattern)

    res = ["(?:^|.+/)" if not anchored else ""]
    while i < n:
        c = pattern[i]
        i += 1
        if c == "*":
            if i < n and pattern[i] == "*":
                i += 1
                if i < n and pattern[i] == "/":
                    i += 1
                    res.append("(.+/)?")  # `/**/` matches `/`.

                else:
                    res.append(".*")

            else:
                res.append("[^/]*")

        elif c == "?":
            res.append("[^/]")

        elif c == "[":
            j = i
            if j < n and pattern[j] == "!":
                j += 1
            if j < n and pattern[j] == "]":
                j += 1
            while j < n and pattern[j] != "]":
                j += 1

            if j >= n:
                res.append("\\[")
            else:
                stuff = pattern[i:j].replace("\\", "\\\\")
                i = j + 1
                if stuff[0] == "!":
                    stuff = f"^{stuff[1:]}"
                elif stuff[0] == "^":
                    stuff = f"\\{stuff}"
                res.append(f"[{stuff}]")

        else:
            res.append(re.escape(c))

    if (
        directory_only
    ):  # In this case we are interested if there is something after slash.
        res.append("(/.+)?$")

    else:
        res.append("(?:/.+)?$")

    return "".join(res)
